#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ribalta', 'ribalta.templates', 'ribalta.utils']

package_data = \
{'': ['*']}

package_dir = \
{'': 'src'}

install_requires = \
['mako >=1.1.3', 'unidecode']

extras_require = \
{'test': ['pytest >=2.7.3', 'pytest-cov', 'python-dateutil']}

setup(name='ribalta',
      version='0.2.6',
      description='RiBa tools!',
      author='Marco Tosato',
      author_email='marco.tosato@didotech.com',
      url='https://github.com/LibrERP/Pylibs/',
      packages=packages,
      package_data=package_data,
      package_dir=package_dir,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>= 3.6',
     )
